@@ -2,9 +2,8 @@
  *
  *  OBEX Server
  *
- *  Copyright (C) 2007-2008  Nokia Corporation
- *  Copyright (C) 2007-2008  Instituto Nokia de Tecnologia (INdT)
- *  Copyright (C) 2007-2009  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2007-2010  Nokia Corporation
+ *  Copyright (C) 2007-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -42,40 +41,83 @@
 #include <glib.h>
 
 #include <openobex/obex.h>
-#include <openobex/obex_const.h>
 
 #include "logging.h"
 #include "obex.h"
+#include "obex-priv.h"
 #include "dbus.h"
 #include "mimetype.h"
 #include "service.h"
+#include "btio.h"
 
 /* Default MTU's */
 #define DEFAULT_RX_MTU 32767
 #define DEFAULT_TX_MTU 32767
 
 /* Connection ID */
-static guint32 cid = 0x0000;
+static uint32_t cid = 0x0000;
 
 static GSList *sessions = NULL;
 
 typedef struct {
-	guint8  version;
-	guint8  flags;
-	guint16 mtu;
+	uint8_t  version;
+	uint8_t  flags;
+	uint16_t mtu;
 } __attribute__ ((packed)) obex_connect_hdr_t;
 
+static void os_set_response(obex_object_t *obj, int err)
+{
+	uint8_t rsp;
+	uint8_t lastrsp;
+
+	switch (err) {
+	case 0:
+		rsp = OBEX_RSP_CONTINUE;
+		lastrsp = OBEX_RSP_SUCCESS;
+		break;
+	case -EPERM:
+	case -EACCES:
+		rsp = OBEX_RSP_FORBIDDEN;
+		lastrsp = OBEX_RSP_FORBIDDEN;
+		break;
+	case -ENOENT:
+		rsp = OBEX_RSP_NOT_FOUND;
+		lastrsp = OBEX_RSP_NOT_FOUND;
+		break;
+	case -EBADR:
+		rsp = OBEX_RSP_BAD_REQUEST;
+		lastrsp = OBEX_RSP_BAD_REQUEST;
+		break;
+	case -EFAULT:
+		rsp = OBEX_RSP_SERVICE_UNAVAILABLE;
+		lastrsp = OBEX_RSP_SERVICE_UNAVAILABLE;
+		break;
+	case -EINVAL:
+		rsp = OBEX_RSP_NOT_IMPLEMENTED;
+		lastrsp = OBEX_RSP_NOT_IMPLEMENTED;
+		break;
+	case -ENOTEMPTY:
+	case -EEXIST:
+		rsp = OBEX_RSP_PRECONDITION_FAILED;
+		lastrsp = OBEX_RSP_PRECONDITION_FAILED;
+		break;
+	default:
+		rsp = OBEX_RSP_INTERNAL_SERVER_ERROR;
+		lastrsp = OBEX_RSP_INTERNAL_SERVER_ERROR;
+	}
+
+	OBEX_ObjectSetRsp(obj, rsp, lastrsp);
+}
+
 static void os_reset_session(struct obex_session *os)
 {
 	if (os->object) {
+		os->driver->set_io_watch(os->object, NULL, NULL);
 		os->driver->close(os->object);
 		os->object = NULL;
-		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->current_folder) {
-			gchar *path;
-			path = g_build_filename(os->current_folder, os->name, NULL);
-			os->driver->remove(path);
-			g_free(path);
-		}
+		os->obj = NULL;
+		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->path)
+			os->driver->remove(os->path);
 	}
 
 	if (os->name) {
@@ -90,6 +132,7 @@ static void os_reset_session(struct obex_session *os)
 		g_free(os->buf);
 		os->buf = NULL;
 	}
+
 	os->driver = NULL;
 	os->aborted = FALSE;
 	os->offset = 0;
@@ -113,9 +156,6 @@ static void obex_session_free(struct obex_session *os)
 
 	os_reset_session(os);
 
-	if (os->current_folder)
-		g_free(os->current_folder);
-
 	if (os->io)
 		g_io_channel_unref(os->io);
 
@@ -123,12 +163,12 @@ static void obex_session_free(struct obex_session *os)
 }
 
 /* From Imendio's GnomeVFS OBEX module (om-utils.c) */
-static time_t parse_iso8610(const gchar *val, int size)
+static time_t parse_iso8610(const char *val, int size)
 {
 	time_t time, tz_offset = 0;
 	struct tm tm;
-	gchar *date;
-	gchar tz;
+	char *date;
+	char tz;
 	int nr;
 
 	memset(&tm, 0, sizeof(tm));
@@ -175,11 +215,12 @@ static void cmd_connect(struct obex_session *os,
 	obex_connect_hdr_t *nonhdr;
 	obex_headerdata_t hd;
 	uint8_t *buffer;
-	guint hlen, newsize;
-	guint16 mtu;
-	guint8 hi;
-	const guint8 *target = NULL, *who = NULL;
-	guint target_size = 0, who_size = 0;
+	unsigned int hlen, newsize;
+	uint16_t mtu;
+	uint8_t hi;
+	const uint8_t *target = NULL, *who = NULL;
+	unsigned int target_size = 0, who_size = 0;
+	int err;
 
 	if (OBEX_ObjectGetNonHdrData(obj, &buffer) != sizeof(*nonhdr)) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -226,16 +267,32 @@ static void cmd_connect(struct obex_session *os,
 
 	debug("Selected driver: %s", os->service->name);
 
-	if (os->service->connect)
-		os->service->connect(obex, obj);
+	if (!os->service->connect) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		return;
+	}
+
+	os->service_data = os->service->connect(os, &err);
+	if (err == 0 && os->service->target) {
+		hd.bs = os->service->target;
+		OBEX_ObjectAddHeader(obex, obj,
+				OBEX_HDR_WHO, hd, 16,
+				OBEX_FL_FIT_ONE_PACKET);
+		hd.bq4 = os->cid;
+		OBEX_ObjectAddHeader(obex, obj,
+				OBEX_HDR_CONNECTION, hd, 4,
+				OBEX_FL_FIT_ONE_PACKET);
+	}
+
+	os_set_response(obj, err);
 }
 
-static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
+static gboolean chk_cid(obex_t *obex, obex_object_t *obj, uint32_t cid)
 {
 	struct obex_session *os;
 	obex_headerdata_t hd;
-	guint hlen;
-	guint8 hi;
+	unsigned int hlen;
+	uint8_t hi;
 	gboolean ret = FALSE;
 
 	os = OBEX_GetUserData(obex);
@@ -260,11 +317,176 @@ static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
 	return ret;
 }
 
+static int obex_read_stream(struct obex_session *os, obex_t *obex,
+						obex_object_t *obj)
+{
+	int size;
+	int32_t len = 0;
+	const uint8_t *buffer;
+
+	if (os->aborted)
+		return -EPERM;
+
+	/* workaround: client didn't send the object lenght */
+	if (os->size == OBJECT_SIZE_DELETE)
+		os->size = OBJECT_SIZE_UNKNOWN;
+
+	if (os->pending > 0)
+		goto write;
+
+	size = OBEX_ObjectReadStream(obex, obj, &buffer);
+	if (size < 0) {
+		error("Error on OBEX stream");
+		return -EIO;
+	}
+
+	if (size > os->rx_mtu) {
+		error("Received more data than RX_MAX");
+		return -EIO;
+	}
+
+	os->buf = g_realloc(os->buf, os->pending + size);
+	memcpy(os->buf + os->pending, buffer, size);
+	os->pending += size;
+	if (os->object == NULL) {
+		debug("Stored %u bytes into temporary buffer", os->pending);
+		return 0;
+	}
+
+write:
+	while (os->pending > 0) {
+		int w;
+
+		w = os->driver->write(os->object, os->buf + len,
+					os->pending);
+		if (w < 0) {
+			if (w == -EINTR)
+				continue;
+			else {
+				memmove(os->buf, os->buf + len, os->pending);
+				return w;
+			}
+		}
+
+		len += w;
+		os->offset += w;
+		os->pending -= w;
+	}
+
+	return 0;
+}
+
+static int obex_write_stream(struct obex_session *os,
+			obex_t *obex, obex_object_t *obj)
+{
+	obex_headerdata_t hd;
+	uint8_t *ptr;
+	int32_t len;
+	unsigned int flags;
+	uint8_t hi;
+
+	debug("obex_write_stream: name=%s type=%s tx_mtu=%d file=%p",
+		os->name ? os->name : "", os->type ? os->type : "",
+		os->tx_mtu, os->object);
+
+	if (os->aborted)
+		return -EPERM;
+
+	if (os->object == NULL) {
+		if (os->buf == NULL && os->finished == FALSE)
+			return -EIO;
+
+		len = MIN(os->size - os->offset, os->tx_mtu);
+		ptr = os->buf + os->offset;
+		goto add_header;
+	}
+
+	len = os->driver->read(os->object, os->buf, os->tx_mtu, &hi);
+	if (len < 0) {
+		error("read(): %s (%d)", strerror(-len), -len);
+		if (len == -EAGAIN)
+			return len;
+		else if (len == -ENOSTR)
+			return 0;
+
+		g_free(os->buf);
+		os->buf = NULL;
+		return len;
+	}
+
+	ptr = os->buf;
+
+add_header:
+
+	hd.bs = ptr;
+
+	switch (hi) {
+	case OBEX_HDR_BODY:
+		flags = len ? OBEX_FL_STREAM_DATA : OBEX_FL_STREAM_DATAEND;
+		break;
+	case OBEX_HDR_APPARAM:
+		flags =  0;
+		break;
+	}
+
+	OBEX_ObjectAddHeader(obex, obj, hi, hd, len, flags);
+
+	if (len == 0) {
+		g_free(os->buf);
+		os->buf = NULL;
+		return len;
+	}
+
+	os->offset += len;
+
+	return len;
+}
+
+static gboolean handle_async_io(void *object, int flags, int err,
+						void *user_data)
+{
+	struct obex_session *os = user_data;
+	int ret = 0;
+
+	if (err < 0) {
+		ret = err;
+		goto proceed;
+	}
+
+	if (flags & (G_IO_IN | G_IO_PRI))
+		ret = obex_write_stream(os, os->obex, os->obj);
+	else if (flags & G_IO_OUT)
+		ret = obex_read_stream(os, os->obex, os->obj);
+
+proceed:
+	switch (ret) {
+	case -EINVAL:
+		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_BAD_REQUEST,
+				OBEX_RSP_BAD_REQUEST);
+	case -EPERM:
+		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_FORBIDDEN,
+					OBEX_RSP_FORBIDDEN);
+		break;
+	default:
+		if (ret < 0)
+			OBEX_ObjectSetRsp(os->obj,
+					OBEX_RSP_INTERNAL_SERVER_ERROR,
+					OBEX_RSP_INTERNAL_SERVER_ERROR);
+		break;
+	}
+
+	OBEX_ResumeRequest(os->obex);
+
+	return FALSE;
+}
+
 static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 {
 	obex_headerdata_t hd;
-	guint hlen;
-	guint8 hi;
+	gboolean stream;
+	unsigned int hlen;
+	uint8_t hi;
+	int err;
 
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -288,7 +510,7 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			if (hlen == 0)
 				continue;
 
-			os->name = g_convert((const gchar *) hd.bs, hlen,
+			os->name = g_convert((const char *) hd.bs, hlen,
 					"UTF8", "UTF16BE", NULL, NULL, NULL);
 			debug("OBEX_HDR_NAME: %s", os->name);
 			break;
@@ -305,39 +527,79 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			if (hd.bs[hlen - 1] != '\0')
 				break;
 
-			if (!g_utf8_validate((const gchar *) hd.bs, -1, NULL)) {
+			if (!g_utf8_validate((const char *) hd.bs, -1, NULL)) {
 				debug("Invalid type header: %s", hd.bs);
 				break;
 			}
 
 			/* FIXME: x-obex/folder-listing - type is mandatory */
 
-			os->type = g_strndup((const gchar *) hd.bs, hlen);
+			os->type = g_strndup((const char *) hd.bs, hlen);
 			debug("OBEX_HDR_TYPE: %s", os->type);
-			os->driver = obex_mime_type_driver_find(os->service->target, os->type);
+			os->driver = obex_mime_type_driver_find(
+						os->service->target, os->type,
+						os->service->who,
+						os->service->who_size);
 			break;
 		}
 	}
 
 	if (!os->driver) {
-		os->driver = obex_mime_type_driver_find(os->service->target, NULL);
+		/* Fallback to target default */
+		os->driver = obex_mime_type_driver_find(os->service->target,
+							NULL, NULL, 0);
 		if (!os->driver) {
-			error("No driver found");
-			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
-					OBEX_RSP_NOT_IMPLEMENTED);
-			return;
+			/* Fallback to general default */
+			os->driver = obex_mime_type_driver_find(NULL,
+							NULL, NULL, 0);
+			if (!os->driver) {
+				error("No driver found");
+				OBEX_ObjectSetRsp(obj,
+						OBEX_RSP_NOT_IMPLEMENTED,
+						OBEX_RSP_NOT_IMPLEMENTED);
+				return;
+			}
 		}
 	}
 
-	os->service->get(obex, obj);
+	err = os->service->get(os, obj, &stream, os->service_data);
+
+	if (err < 0)
+		goto done;
+
+	if (os->size != OBJECT_SIZE_UNKNOWN) {
+		hd.bq4 = os->size;
+		OBEX_ObjectAddHeader(obex, obj,
+				OBEX_HDR_LENGTH, hd, 4, 0);
+	}
+
+	/* Add body header */
+	hd.bs = NULL;
+	if (os->size == 0)
+		OBEX_ObjectAddHeader (obex, obj, OBEX_HDR_BODY,
+				hd, 0, OBEX_FL_FIT_ONE_PACKET);
+	else if (!stream) {
+		/* Asynchronous operation that doesn't use stream */
+		OBEX_SuspendRequest(obex, obj);
+		os->obj = obj;
+		os->driver->set_io_watch(os->object, handle_async_io, os);
+		return;
+	} else
+		/* Standard data stream */
+		OBEX_ObjectAddHeader (obex, obj, OBEX_HDR_BODY,
+				hd, 0, OBEX_FL_STREAM_START);
+
+done:
+	os_set_response(obj, err);
 }
 
 static void cmd_setpath(struct obex_session *os,
 			obex_t *obex, obex_object_t *obj)
 {
 	obex_headerdata_t hd;
-	guint32 hlen;
-	guint8 hi;
+	uint32_t hlen;
+	int err;
+	uint8_t hi;
 
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -371,32 +633,35 @@ static void cmd_setpath(struct obex_session *os,
 			break;
 		}
 
-		os->name = g_convert((const gchar *) hd.bs, hlen,
+		os->name = g_convert((const char *) hd.bs, hlen,
 				"UTF8", "UTF16BE", NULL, NULL, NULL);
 
 		debug("Set path name: %s", os->name);
 		break;
 	}
 
-	os->service->setpath(obex, obj);
+	err = os->service->setpath(os, obj, os->service_data);
+	os_set_response(obj, err);
 }
 
-int os_prepare_get(struct obex_session *os, gchar *filename, size_t *size)
+int obex_get_stream_start(struct obex_session *os, const char *filename)
 {
-	gint err;
-	gpointer object;
+	int err;
+	void *object;
+	size_t size;
 
-	object = os->driver->open(filename, O_RDONLY, 0, size);
+	object = os->driver->open(filename, O_RDONLY, 0, os->service_data,
+								&size, &err);
 	if (object == NULL) {
-		err = -errno;
-		error("open(%s): %s (%d)", filename, strerror(errno), errno);
+		error("open(%s): %s (%d)", filename, strerror(-err), -err);
 		goto fail;
 	}
 
 	os->object = object;
 	os->offset = 0;
+	os->size = size;
 
-	if (*size > 0)
+	if (size > 0)
 		os->buf = g_malloc0(os->tx_mtu);
 
 	return 0;
@@ -404,167 +669,42 @@ int os_prepare_get(struct obex_session *os, gchar *filename, size_t *size)
 fail:
 	if (object)
 		os->driver->close(object);
+
 	return err;
 }
 
-static gint obex_write_stream(struct obex_session *os,
-			obex_t *obex, obex_object_t *obj)
+int obex_put_stream_start(struct obex_session *os, const char *filename)
 {
-	obex_headerdata_t hd;
-	gint32 len;
-	guint8 *ptr;
-
-	debug("obex_write_stream: name=%s type=%s tx_mtu=%d file=%p",
-		os->name ? os->name : "", os->type ? os->type : "",
-		os->tx_mtu, os->object);
-
-	if (os->aborted)
-		return -EPERM;
+	int err;
 
+	os->object = os->driver->open(filename, O_WRONLY | O_CREAT | O_TRUNC,
+					0600, os->service_data,
+					os->size != OBJECT_SIZE_UNKNOWN ?
+					(size_t *) &os->size : NULL, &err);
 	if (os->object == NULL) {
-		if (os->buf == NULL && os->finished == FALSE)
-			return -EIO;
-
-		len = MIN(os->size - os->offset, os->tx_mtu);
-		ptr = os->buf + os->offset;
-		goto add_header;
-	}
-
-	len = os->driver->read(os->object, os->buf, os->tx_mtu);
-	if (len < 0) {
-		gint err = errno;
-		error("read(): %s (%d)", strerror(err), err);
-		g_free(os->buf);
-		os->buf = NULL;
-		return -err;
-	}
-
-	ptr = os->buf;
-
-add_header:
-
-	hd.bs = ptr;
-
-	if (len == 0) {
-		OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, 0,
-					OBEX_FL_STREAM_DATAEND);
-		g_free(os->buf);
-		os->buf = NULL;
-		return len;
-	}
-
-	os->offset += len;
-
-	OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, len,
-				OBEX_FL_STREAM_DATA);
-
-	return len;
-}
-
-gint os_prepare_put(struct obex_session *os)
-{
-	gchar *path;
-	gint len;
-
-	path = g_build_filename(os->current_folder, os->name, NULL);
-
-	os->object = os->driver->open(path, O_WRONLY | O_CREAT | O_TRUNC, 0600,
-					(size_t *) &os->size);
-	if (os->object == NULL) {
-		error("open(%s): %s (%d)", path, strerror(errno), errno);
-		g_free(path);
+		error("open(%s): %s (%d)", filename, strerror(-err), -err);
 		return -EPERM;
 	}
 
-	g_free(path);
+	os->path = filename;
 
 	if (!os->buf) {
 		debug("PUT request checked, no buffered data");
 		return 0;
 	}
 
-	len = 0;
-	while (len < os->offset) {
-		gint w;
-
-		w = os->driver->write(os->object, os->buf + len, os->offset - len);
-		if (w < 0) {
-			gint err = errno;
-			error("write(%s): %s (%d)", path, strerror(errno),
-									errno);
-			if (err == EINTR)
-				continue;
-			else
-				return -err;
-		}
-
-		len += w;
-	}
-
-	return 0;
-}
-
-static gint obex_read_stream(struct obex_session *os, obex_t *obex,
-				obex_object_t *obj)
-{
-	gint size;
-	gint32 len = 0;
-	const guint8 *buffer;
-
-	if (os->aborted)
-		return -EPERM;
-
-	/* workaround: client didn't send the object lenght */
-	if (os->size == OBJECT_SIZE_DELETE)
-		os->size = OBJECT_SIZE_UNKNOWN;
-
-	size = OBEX_ObjectReadStream(obex, obj, &buffer);
-	if (size < 0) {
-		error("Error on OBEX stream");
-		return -EIO;
-	}
-
-	if (size > os->rx_mtu) {
-		error("Received more data than RX_MAX");
-		return -EIO;
-	}
-
-	if (os->object == NULL && size > 0) {
-		os->buf = g_realloc(os->buf, os->offset + size);
-		memcpy(os->buf + os->offset, buffer, size);
-		os->offset += size;
-
-		debug("Stored %u bytes into temporary buffer", size);
-
+	if (os->pending == 0)
 		return 0;
-	}
 
-	while (len < size) {
-		gint w;
-
-		w = os->driver->write(os->object, buffer + len, size - len);
-		if (w < 0) {
-			gint err = errno;
-			if (err == EINTR)
-				continue;
-			else
-				return -err;
-		}
-
-		len += w;
-	}
-
-	os->offset += len;
-
-	return 0;
+	return obex_read_stream(os, os->obex, NULL);
 }
 
 static gboolean check_put(obex_t *obex, obex_object_t *obj)
 {
 	struct obex_session *os;
 	obex_headerdata_t hd;
-	guint hlen;
-	guint8 hi;
+	unsigned int hlen;
+	uint8_t hi;
 	int ret;
 
 	os = OBEX_GetUserData(obex);
@@ -590,7 +730,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 			if (hlen == 0)
 				continue;
 
-			os->name = g_convert((const gchar *) hd.bs, hlen,
+			os->name = g_convert((const char *) hd.bs, hlen,
 					"UTF8", "UTF16BE", NULL, NULL, NULL);
 			debug("OBEX_HDR_NAME: %s", os->name);
 			break;
@@ -608,14 +748,17 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 			if (hd.bs[hlen - 1] != '\0')
 				break;
 
-			if (!g_utf8_validate((const gchar *) hd.bs, -1, NULL)) {
+			if (!g_utf8_validate((const char *) hd.bs, -1, NULL)) {
 				debug("Invalid type header: %s", hd.bs);
 				break;
 			}
 
-			os->type = g_strndup((const gchar *) hd.bs, hlen);
+			os->type = g_strndup((const char *) hd.bs, hlen);
 			debug("OBEX_HDR_TYPE: %s", os->type);
-			os->driver = obex_mime_type_driver_find(os->service->target, os->type);
+			os->driver = obex_mime_type_driver_find(
+						os->service->target, os->type,
+						os->service->who,
+						os->service->who_size);
 			break;
 
 		case OBEX_HDR_BODY:
@@ -628,7 +771,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 			debug("OBEX_HDR_LENGTH: %d", os->size);
 			break;
 		case OBEX_HDR_TIME:
-			os->time = parse_iso8610((const gchar *) hd.bs, hlen);
+			os->time = parse_iso8610((const char *) hd.bs, hlen);
 			break;
 		}
 	}
@@ -636,34 +779,48 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 	OBEX_ObjectReParseHeaders(obex, obj);
 
 	if (!os->driver) {
-		os->driver = obex_mime_type_driver_find(os->service->target, NULL);
+		/* Fallback to target default */
+		os->driver = obex_mime_type_driver_find(os->service->target,
+							NULL, NULL, 0);
 		if (!os->driver) {
-			error("No driver found");
-			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
-					OBEX_RSP_NOT_IMPLEMENTED);
-			return FALSE;
+			/* Fallback to general default */
+			os->driver = obex_mime_type_driver_find(NULL,
+							NULL, NULL, 0);
+			if (!os->driver) {
+				error("No driver found");
+				OBEX_ObjectSetRsp(obj,
+						OBEX_RSP_NOT_IMPLEMENTED,
+						OBEX_RSP_NOT_IMPLEMENTED);
+				return FALSE;
+			}
 		}
 	}
 
-	if (!os->service || !os->service->chkput)
+	if (!os->service->chkput)
 		goto done;
 
-	ret = os->service->chkput(obex, obj);
+	ret = os->service->chkput(os, os->service_data);
 	switch (ret) {
 	case 0:
 		break;
-	case -EINVAL:
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
-				OBEX_RSP_BAD_REQUEST);
-		return FALSE;
 	case -EPERM:
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return FALSE;
+	case -EBADR:
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
+					OBEX_RSP_BAD_REQUEST);
+		return FALSE;
+	case -EAGAIN:
+		OBEX_SuspendRequest(obex, obj);
+		os->obj = obj;
+		os->driver->set_io_watch(os->object, handle_async_io, os);
+		return TRUE;
 	default:
 		debug("Unhandled chkput error: %d", ret);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_INTERNAL_SERVER_ERROR,
 				OBEX_RSP_INTERNAL_SERVER_ERROR);
 		return FALSE;
+
 	}
 
 	if (os->size == OBJECT_SIZE_DELETE || os->size == OBJECT_SIZE_UNKNOWN) {
@@ -679,6 +836,8 @@ done:
 
 static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 {
+	int err;
+
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
@@ -695,11 +854,13 @@ static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			return;
 	}
 
-	os->service->put(obex, obj);
+	err = os->service->put(os, os->service_data);
+	if (err < 0)
+		os_set_response(obj, err);
 }
 
-static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
-					gint evt, gint cmd, gint rsp)
+static void obex_event(obex_t *obex, obex_object_t *obj, int mode,
+					int evt, int cmd, int rsp)
 {
 	struct obex_session *os;
 
@@ -710,30 +871,31 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	switch (evt) {
 	case OBEX_EV_PROGRESS:
 		if (os->service->progress)
-			os->service->progress(obex, obj);
+			os->service->progress(os, os->service_data);
 		break;
 	case OBEX_EV_ABORT:
 		os->aborted = TRUE;
 		if (os->service->reset)
-			os->service->reset(obex);
+			os->service->reset(os, os->service_data);
 		os_reset_session(os);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
 		break;
 	case OBEX_EV_REQDONE:
 		switch (cmd) {
+		case OBEX_CMD_CONNECT:
+			break;
 		case OBEX_CMD_DISCONNECT:
 			OBEX_TransportDisconnect(obex);
 			break;
 		case OBEX_CMD_PUT:
 		case OBEX_CMD_GET:
 		case OBEX_CMD_SETPATH:
+		default:
 			os_session_mark_aborted(os);
 			if (os->service->reset)
-				os->service->reset(obex);
+				os->service->reset(os, os->service_data);
 			os_reset_session(os);
 			break;
-		default:
-			break;
 		}
 		break;
 	case OBEX_EV_REQHINT:
@@ -758,7 +920,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	case OBEX_EV_REQCHECK:
 		switch (cmd) {
 		case OBEX_CMD_PUT:
-			if (os->service && os->service->put)
+			if (os->service)
 				check_put(obex, obj);
 			break;
 		default:
@@ -783,8 +945,8 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 			break;
 		default:
 			debug("Unknown request: 0x%X", cmd);
-			OBEX_ObjectSetRsp(obj,
-				OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+						OBEX_RSP_NOT_IMPLEMENTED);
 			break;
 		}
 		break;
@@ -796,6 +958,12 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 			OBEX_ObjectSetRsp(obj,
 				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 			break;
+		case -EAGAIN:
+			OBEX_SuspendRequest(obex, obj);
+			os->obj = obj;
+			os->driver->set_io_watch(os->object, handle_async_io,
+									os);
+			break;
 		default:
 			OBEX_ObjectSetRsp(obj,
 				OBEX_RSP_INTERNAL_SERVER_ERROR,
@@ -805,7 +973,21 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 
 		break;
 	case OBEX_EV_STREAMEMPTY:
-		obex_write_stream(os, obex, obj);
+		switch (obex_write_stream(os, obex, obj)) {
+		case -EPERM:
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN,
+							OBEX_RSP_FORBIDDEN);
+			break;
+		case -EAGAIN:
+			OBEX_SuspendRequest(obex, obj);
+			os->obj = obj;
+			os->driver->set_io_watch(os->object, handle_async_io,
+									os);
+			break;
+		default:
+			break;
+		}
+
 		break;
 	case OBEX_EV_LINKERR:
 		break;
@@ -828,7 +1010,7 @@ void server_free(struct server *server)
 	g_free(server);
 }
 
-static void obex_handle_destroy(gpointer user_data)
+static void obex_handle_destroy(void *user_data)
 {
 	struct obex_session *os;
 	obex_t *obex = user_data;
@@ -836,18 +1018,18 @@ static void obex_handle_destroy(gpointer user_data)
 	os = OBEX_GetUserData(obex);
 
 	if (os->service && os->service->disconnect)
-		os->service->disconnect(obex);
+		os->service->disconnect(os, os->service_data);
 
 	obex_session_free(os);
 
 	OBEX_Cleanup(obex);
 }
 
-static gboolean tty_reinit(gpointer data)
+static gboolean tty_reinit(void *data)
 {
 	struct server *server = data;
 	GSList *l;
-	guint services = 0;
+	unsigned int services = 0;
 
 	for (l = server->drivers; l; l = l->next) {
 		struct obex_service_driver *driver = l->data;
@@ -864,7 +1046,7 @@ static gboolean tty_reinit(gpointer data)
 }
 
 static gboolean obex_handle_input(GIOChannel *io,
-				GIOCondition cond, gpointer user_data)
+				GIOCondition cond, void *user_data)
 {
 	obex_t *obex = user_data;
 	struct obex_session *os = OBEX_GetUserData(obex);
@@ -895,17 +1077,30 @@ failed:
 	return FALSE;
 }
 
-gint obex_session_start(GIOChannel *io, struct server *server)
+void obex_connect_cb(GIOChannel *io, GError *err, void *user_data)
+{
+	struct server *server = user_data;
+
+	if (err) {
+		error("%s", err->message);
+		g_io_channel_shutdown(io, TRUE, NULL);
+		return;
+	}
+
+	if (obex_session_start(io, server) < 0)
+		g_io_channel_shutdown(io, TRUE, NULL);
+}
+
+int obex_session_start(GIOChannel *io, struct server *server)
 {
 	struct obex_session *os;
 	obex_t *obex;
-	gint ret, fd;
+	int ret, fd;
 
 	os = g_new0(struct obex_session, 1);
 
 	os->service = obex_service_driver_find(server->drivers, NULL, 0,
 						NULL, 0);
-	os->current_folder = g_strdup(server->folder);
 	os->server = server;
 	os->rx_mtu = server->rx_mtu ? server->rx_mtu : DEFAULT_RX_MTU;
 	os->tx_mtu = server->tx_mtu ? server->tx_mtu : DEFAULT_TX_MTU;
@@ -941,7 +1136,7 @@ gint obex_session_start(GIOChannel *io, struct server *server)
 	return 0;
 }
 
-gint obex_tty_session_stop(void)
+int obex_tty_session_stop(void)
 {
 	GSList *l;
 
@@ -955,16 +1150,108 @@ gint obex_tty_session_stop(void)
 	return 0;
 }
 
-struct obex_session *obex_get_session(gpointer object)
+const char *obex_get_name(struct obex_session *os)
 {
-	GSList *l;
+	return os->name;
+}
 
-	for (l = sessions; l; l = l->next) {
-		struct obex_session *os = l->data;
+ssize_t obex_get_size(struct obex_session *os)
+{
+	return os->size;
+}
 
-		if (os->object == object)
-			return os;
+const char *obex_get_type(struct obex_session *os)
+{
+	return os->type;
+}
+
+const char *obex_get_root_folder(struct obex_session *os)
+{
+	return os->server->folder;
+}
+
+uint16_t obex_get_service(struct obex_session *os)
+{
+	return os->service->service;
+}
+
+gboolean obex_get_symlinks(struct obex_session *os)
+{
+	return os->server->symlinks;
+}
+
+const char *obex_get_capability_path(struct obex_session *os)
+{
+	return os->server->capability;
+}
+
+gboolean obex_get_auto_accept(struct obex_session *os)
+{
+	return os->server->auto_accept;
+}
+
+int obex_remove(struct obex_session *os, const char *path)
+{
+	if (os->driver == NULL)
+		return -EINVAL;
+
+	return os->driver->remove(path);
+}
+
+/* TODO: find a way to do this for tty or fix syncevolution */
+char *obex_get_id(struct obex_session *os)
+{
+	GError *gerr = NULL;
+	char address[18];
+	uint8_t channel;
+
+	bt_io_get(os->io, BT_IO_RFCOMM, &gerr,
+			BT_IO_OPT_DEST, address,
+			BT_IO_OPT_CHANNEL, &channel,
+			BT_IO_OPT_INVALID);
+	if (gerr)
+		return NULL;
+
+	return g_strdup_printf("%s+%d", address, channel);
+}
+
+ssize_t obex_aparam_read(struct obex_session *os,
+		obex_object_t *obj, const uint8_t **buffer)
+{
+	obex_headerdata_t hd;
+	uint8_t hi;
+	uint32_t hlen;
+
+	OBEX_ObjectReParseHeaders(os->obex, obj);
+
+	while (OBEX_ObjectGetNextHeader(os->obex, obj, &hi, &hd, &hlen)) {
+		if (hi == OBEX_HDR_APPARAM) {
+			*buffer = hd.bs;
+			return hlen;
+		}
 	}
 
-	return NULL;
+	return -EBADR;
+}
+
+int obex_aparam_write(struct obex_session *os,
+		obex_object_t *obj, const uint8_t *data, unsigned int size)
+{
+	obex_headerdata_t hd;
+
+	hd.bs = data;
+
+	return OBEX_ObjectAddHeader(os->obex, obj,
+			OBEX_HDR_APPARAM, hd, size, 0);
+}
+
+int memcmp0(const void *a, const void *b, size_t n)
+{
+	if (a == NULL)
+		return -(a != b);
+
+	if (b == NULL)
+		return a != b;
+
+	return memcmp(a, b, n);
 }
